package ch.unige.solidify.test.helper;

import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import org.apache.tomcat.util.buf.StringUtils;
import org.json.JSONObject;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.snippet.Attributes.Attribute;

import ch.unige.solidify.test.exception.SolidifyTestRuntimeException;
import ch.unige.solidify.test.model.OpenApiExtraDocumentation;
import ch.unige.solidify.test.model.OpenApiExtraDocumentationProperties;
import ch.unige.solidify.test.model.TypeField;
import ch.unige.solidify.util.StringTool;

public abstract class AbstractResourceDoc {

  public static final String CONSTRAINTS = "constraints";
  public static final String ENUM_VALUE_QUOTE = "`";
  public static final char ENUM_VALUE_SEPARATOR = ',';
  public static final String FORMAT_INTEGER_INT32 = "int32";
  public static final String FORMAT_INTEGER_INT64 = "int64";
  public static final String FORMAT_NUMBER_DOUBLE = "double";
  public static final String FORMAT_NUMBER_FLOAT = "float";
  public static final String FORMAT_PREFIX = "_Format_:";

  public static final String FORMAT_STRING_BINARY = "binary";
  public static final String FORMAT_STRING_BYTE = "byte";
  public static final String FORMAT_STRING_DATE = "date";
  public static final String FORMAT_STRING_DATE_TIME = "date-time";
  @SuppressWarnings("squid:S2068")
  public static final String FORMAT_STRING_PWD = "password";
  public static final String MANDATORY_VALUE = "mandatoryValue";
  public static final String NO_CONSTRAINT = "";
  public static final String READ_ONLY_VALUE = "readOnlyValue";
  public static final String TEST_STRING_VALUE = "test";
  public static final String VALUE_PREFIX = "_Values_:";
  protected static final String CONSTRAINT_CODE_ISO = "code ISO 3166-2";

  protected static final String CONSTRAINT_DATE_INTEGER = "3650 days ~ 10 years, 0 day ~ forever";
  protected static final String CONSTRAINT_VALID_DOI = "Valid DOI";
  protected static final String CONSTRAINT_VALID_ID = "Valid ID";
  protected static final String CONSTRAINT_VALID_URL = "Valid URL";
  protected static final String CONSTRAINT_VALID_JSON = "Valid JSON";
  protected static final String NO_TEST_VALUE = "NO_TEST_VALUE";

  protected static final String TEST_CODE_ISO = "CH";
  // Test value list
  protected static final String TEST_DATE_FORMAT_VALUE = "2018-07-10T15:36:50.323+0200";
  protected static final String TEST_DATE_SIMPLE_VALUE = "2018-07-10";
  protected static final String TEST_EMAIL_VALUE = "solidify@unige.ch";
  protected static final String TEST_NUMBER_VALUE = "5";
  protected static final String TEST_ORCID = "0000-1111-2222-3333";
  protected static final String TEST_SEARCH_QUERY = "*";
  protected static final String TEST_TRUE_VALUE = "true";
  protected static final String TEST_URL_VALUE = "https://www.unige.ch/";
  protected static final String TEST_JSON_OBJECT_VALUE = "{ \"test\" : \"value\" }";
  protected static final String TEST_JSON_ARRAY_VALUE = "[{ \"test\" : \"value\" }, { \"test2\" : \"value2\" }]";

  public static String getCamelCase(String className) {
    return className.substring(0, 1).toLowerCase().concat(className.substring(1));
  }

  private final List<FieldDescriptor> docFields;

  private final OpenApiExtraDocumentation openApiExtraDocumentation;

  protected AbstractResourceDoc() {
    this.openApiExtraDocumentation = new OpenApiExtraDocumentation(this.getName(), this.getModelName(), this.isNormalResource());
    this.docFields = this.getDocumentationFields();
    this.sortFields();
  }

  public JSONObject getJsonFields() {
    return TestObjectBuilder.buildJsonObject(this.docFields);
  }

  public abstract List<String> getListLinks();

  public String getModelName() {
    if (this.getName().lastIndexOf('/') == -1) {
      return null;
    }
    final String resource = this.getName().substring(this.getName().lastIndexOf('/') + 1);
    if (resource.endsWith("ies")) {
      return resource.substring(0, resource.length() - 3) + "y";
    } else if (resource.endsWith("s")) {
      return resource.substring(0, resource.length() - 1);
    }
    return resource;
  }

  public abstract String getName();

  public int getNumberOfExpectedReadOnlyFields() {
    final Set<String> firstLevelPathSet = new HashSet<>();
    for (final FieldDescriptor fieldDescriptor : this.docFields) {
      if (!fieldDescriptor.isOptional() && String.valueOf(true).equals(fieldDescriptor.getAttributes().get(READ_ONLY_VALUE))) {
        final String firstLevelPath = fieldDescriptor.getPath().split("\\.")[0];
        firstLevelPathSet.add(firstLevelPath);
      }
    }
    return firstLevelPathSet.size();
  }

  public OpenApiExtraDocumentation getOpenApiExtraDocumentation() {
    return this.openApiExtraDocumentation;
  }

  public List<FieldDescriptor> getReadOnlyDocumentationFields() {
    return this.docFields.stream().filter(
            field -> field.getAttributes().get(READ_ONLY_VALUE).equals(String.valueOf(true)))
            .collect(Collectors.toList());
  }

  public List<String> getCommonResourceFields() {
    return Collections.unmodifiableList(Arrays.asList(RestDocsTemplates.RESOURCE_ID_FIELD, RestDocsTemplates.CREATION_FIELD,
            RestDocsTemplates.LAST_UPDATE_FIELD, RestDocsTemplates.LINKS_FIELD));
  }

  public boolean isNormalResource() {
    // By default
    return true;
  }

  public abstract List<String> getResourceLinks();

  protected FieldDescriptor createFieldDescriptor(String name, String description) {
    return this.createFieldDescriptor(name, description, false, TypeField.UNDEFINED, null, null, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, List<String> enumValues) {
    return this.createFieldDescriptor(name, description, mandatory, TypeField.STRING, enumValues, null, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, List<String> enumValues,
          String testValue) {
    return this.createFieldDescriptor(name, description, mandatory, TypeField.STRING, enumValues, null, testValue, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, TypeField typeField) {
    return this.createFieldDescriptor(name, description, mandatory, typeField, null, null, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, TypeField type, List<String> enumValues,
          String customConstraints,
          String testValue, boolean readOnly) {

    final String constraints = this.determineConstraint(customConstraints, type, enumValues);
    testValue = this.determineTestValue(testValue, type, enumValues);
    final FieldDescriptor fieldDescriptor = fieldWithPath(name).description(description);
    this.setFieldDescriptorAttributes(fieldDescriptor, mandatory, testValue, constraints, readOnly);

    this.addToOpenApiExtraDocumentation(name, enumValues, type);
    return fieldDescriptor;
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, TypeField type, String customConstraints) {
    return this.createFieldDescriptor(name, description, mandatory, type, null, customConstraints, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, TypeField type, String customConstraints,
          boolean readOnly) {
    return this.createFieldDescriptor(name, description, mandatory, type, null, customConstraints, null, readOnly);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, boolean mandatory, TypeField type, String customConstraints,
          String testValue) {
    return this.createFieldDescriptor(name, description, mandatory, type, null, customConstraints,
            testValue, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, List<String> enumValues) {
    return this.createFieldDescriptor(name, description, false, TypeField.STRING, enumValues, null, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, List<String> enumValues, String testValue) {
    return this.createFieldDescriptor(name, description, false, TypeField.STRING, enumValues, null, testValue, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, String item) {
    final FieldDescriptor fieldDescriptor = this.createFieldDescriptor(name, description, false,
            TypeField.ARRAY, null, null, null, true);
    this.openApiExtraDocumentation.getProperties()
            .add(new OpenApiExtraDocumentationProperties(name, null, item));
    return fieldDescriptor;
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, TypeField typeField) {
    return this.createFieldDescriptor(name, description, false, typeField, null, null, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, TypeField type, String customConstraints) {
    return this.createFieldDescriptor(name, description, false, type, null, customConstraints, null, true);
  }

  protected FieldDescriptor createFieldDescriptor(String name, String description, TypeField type, String customConstraints, String testValue) {
    return this.createFieldDescriptor(name, description, false, type, null, customConstraints, testValue, true);
  }

  protected String displayFormat(String format) {
    return FORMAT_PREFIX + " " + ENUM_VALUE_QUOTE + format + ENUM_VALUE_QUOTE;
  }

  protected <T> List<String> enumToList(T[] enumValues) {
    return this.enumToList(enumValues, false);
  }

  protected <T> List<String> enumToList(T[] enumValues, boolean toLowerCase) {
    final List<String> listStringEnumValues = new ArrayList<>();
    for (final T value : enumValues) {
      String val = value.toString();
      if (toLowerCase) {
        val = val.toLowerCase();
      }
      listStringEnumValues.add(val);
    }
    return listStringEnumValues;
  }

  protected abstract List<FieldDescriptor> getDocumentationFields();

  public void updateValues(Map<String, String> properties) {
    // Do nothing by default
  }

  protected void updateValues(Map<String, String> properties, String fieldName, String valueName) {
    // Test if value exist
    if (!properties.containsKey(valueName)) {
      throw new SolidifyTestRuntimeException("[" + this.getClass().getSimpleName() + "] Cannot find '" + valueName + "' value in property list");
    }
    for (final FieldDescriptor field : this.docFields) {
      if (field.getPath().equals(fieldName)) {
        field.getAttributes().put(TEST_STRING_VALUE, properties.get(valueName));
        return;
      }
    }
    throw new SolidifyTestRuntimeException("[" + this.getClass().getSimpleName() + "] Cannot find '" + fieldName + "' field in field list");
  }

  private void addToOpenApiExtraDocumentation(String fieldName, List<String> enumValues,
          TypeField type) {
    final String format = this.getExtraDocumentationFormat(type);
    if (format != null || enumValues != null) {
      this.openApiExtraDocumentation.getProperties()
              .add(new OpenApiExtraDocumentationProperties(fieldName, format, enumValues));
    }
  }

  private String determineConstraint(String constraints, TypeField type, List<String> enumValues) {
    if (constraints != null) {
      return constraints;
    }

    if (enumValues != null) {
      return this.listValues(enumValues);
    }

    switch (type) {
      case DATE:
        constraints = this.displayFormat(StringTool.DATE_FORMAT);
        break;
      case DATE_TIME:
        constraints = this.displayFormat(StringTool.DATE_TIME_FORMAT);
        break;
      default:
        constraints = NO_CONSTRAINT;
    }
    return constraints;
  }

  private String determineTestValue(String testValue, TypeField type, List<String> enumValues) {
    if (testValue != null) {
      if (testValue.equals(NO_TEST_VALUE)) {
        return null;
      }
      return testValue;
    }

    if (enumValues != null) {
      if (enumValues.isEmpty()) {
        return null;
      }
      return enumValues.get(0);
    }

    switch (type) {
      case INT32:
      case INT64:
      case FLOAT:
      case DOUBLE:
        testValue = TEST_NUMBER_VALUE;
        break;
      case STRING:
      case BYTE:
      case BINARY:
      case PASSWORD:
        testValue = TEST_STRING_VALUE;
        break;
      case BOOLEAN:
        testValue = TEST_TRUE_VALUE;
        break;
      case DATE:
        testValue = TEST_DATE_SIMPLE_VALUE;
        break;
      case DATE_TIME:
        testValue = TEST_DATE_FORMAT_VALUE;
        break;
      case ARRAY:
      case UNDEFINED:
        break;
      case URL:
        testValue = TEST_URL_VALUE;
        break;
    }
    return testValue;
  }

  private String getExtraDocumentationFormat(TypeField type) {
    String format = null;
    switch (type) {
      case INT32:
        format = FORMAT_INTEGER_INT32;
        break;
      case INT64:
        format = FORMAT_INTEGER_INT64;
        break;
      case FLOAT:
        format = FORMAT_NUMBER_FLOAT;
        break;
      case DOUBLE:
        format = FORMAT_NUMBER_DOUBLE;
        break;
      case BYTE:
        format = FORMAT_STRING_BYTE;
        break;
      case BINARY:
        format = FORMAT_STRING_BINARY;
        break;
      case BOOLEAN:
        break;
      case DATE:
        format = FORMAT_STRING_DATE;
        break;
      case DATE_TIME:
        format = FORMAT_STRING_DATE_TIME;
        break;
      case PASSWORD:
        format = FORMAT_STRING_PWD;
        break;
      case STRING:
      case UNDEFINED:
      case ARRAY:
      case URL:
        break;
      default:
        break;
    }
    return format;
  }

  private String listValues(List<String> enumValues) {
    final List<String> enumValuesFormated = enumValues.stream()
            .map(val -> " " + ENUM_VALUE_QUOTE + val + ENUM_VALUE_QUOTE)
            .collect(Collectors.toList());
    return VALUE_PREFIX + StringUtils.join(enumValuesFormated, ENUM_VALUE_SEPARATOR);
  }

  private void setFieldDescriptorAttributes(FieldDescriptor fieldDescriptor, boolean mandatory,
          String testValue, String constraints, boolean readOnly) {
    fieldDescriptor.attributes(new Attribute(CONSTRAINTS, constraints));
    // Add attribute only true because AsciiDoc test if the value is present
    if (mandatory) {
      fieldDescriptor.attributes(new Attribute(MANDATORY_VALUE, String.valueOf(mandatory)));
    }
    fieldDescriptor.attributes(new Attribute(READ_ONLY_VALUE, String.valueOf(readOnly)));

    this.setTestValue(fieldDescriptor, testValue);
  }

  private void setTestValue(FieldDescriptor fieldDescriptor, String testValue) {
    if (testValue == null) {
      return;
    }
    fieldDescriptor.attributes(new Attribute(TEST_STRING_VALUE, testValue));
  }

  private void sortFields() {
    this.docFields.sort((Comparator.comparing(FieldDescriptor::getPath)));
  }
}
