package ch.unige.solidify.rest;

import java.lang.reflect.Field;
import java.time.OffsetDateTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

import javax.persistence.Transient;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.hateoas.RepresentationModel;

import com.fasterxml.jackson.annotation.JsonIgnore;

import ch.unige.solidify.model.ChangeInfo;
import ch.unige.solidify.util.ReflectionTool;

public abstract class ResourceBase extends RepresentationModel<ResourceBase> {
  private static final Logger logger = LoggerFactory.getLogger(ResourceBase.class);

  @Transient
  @JsonIgnore
  protected List<String> dirtyFields;

  public ResourceBase() {
    this.dirtyFields = new ArrayList<>();
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj) {
      return true;
    }
    if (!(obj instanceof ResourceBase)) {
      return false;
    }
    final ResourceBase resourceBase = (ResourceBase) obj;
    if (this.getCreationTime() != null && this.getUpdateTime() != null) {
      return super.equals(obj)
              && Objects.equals(this.getCreationTime(), resourceBase.getCreationTime())
              && Objects.equals(this.getUpdateTime(), resourceBase.getUpdateTime());
    } else if (this.getCreationTime() != null) {
      return super.equals(obj)
              && Objects.equals(this.getCreationTime(), resourceBase.getCreationTime());
    } else if (this.getUpdateTime() != null) {
      return super.equals(obj)
              && Objects.equals(this.getUpdateTime(), resourceBase.getUpdateTime());
    }
    return super.equals(obj);
  }

  @JsonIgnore
  public abstract String getCreatedBy();

  public abstract ChangeInfo getCreation();

  @JsonIgnore
  @Transient
  public abstract OffsetDateTime getCreationTime();

  @Transient
  public List<String> getDirtyFields() {
    return this.dirtyFields;
  }

  public abstract ChangeInfo getLastUpdate();

  @JsonIgnore
  public abstract String getUpdatedBy();

  @JsonIgnore
  public abstract OffsetDateTime getUpdateTime();

  @Override
  public int hashCode() {
    return Objects.hash(super.hashCode(), this.getCreationTime(), this.getUpdateTime());
  }

  @Transient
  @JsonIgnore
  public boolean isDirty() {
    return !this.dirtyFields.isEmpty();
  }

  @Transient
  public boolean isDirty(String fieldName) {
    return this.dirtyFields.contains(fieldName);
  }

  public abstract void setCreatedBy(String createdBy);

  public abstract void setUpdatedBy(String updatedBy);

  /**
   * Update the current Resource with the dirty attributes of the given Resource
   *
   * @param item
   *          Resource
   * @return boolean Return true if the current Resource has been modified
   */
  public boolean update(ResourceBase item) {
    final OffsetDateTime lastUpdate = this.getUpdateTime();

    final List<String> dirtyFields = item.getDirtyFields();

    for (final String fieldname : dirtyFields) {
      try {
        final Field field = ReflectionTool.getField(this, fieldname);
        field.setAccessible(true);
        final Object newValue = field.get(item);
        final Object currentValue = field.get(this);

        if (!Objects.equals(currentValue, newValue)) {
          this.setProperty(fieldname, newValue);
        }

      } catch (NoSuchFieldException | IllegalAccessException e) {
        return false;
      }
    }

    return (lastUpdate != this.getUpdateTime());
  }

  public abstract void updateTime();

  @Transient
  protected boolean setProperty(String fieldName, Object newValue) {
    if (this.getDirtyFields() == null) {
      this.dirtyFields = new ArrayList<>();
    }

    try {
      final Field field = ReflectionTool.getField(this, fieldName);
      field.setAccessible(true);

      if (newValue instanceof EmbeddableEntity) {
        final EmbeddableEntity currentEmbeddableEntity = (EmbeddableEntity) field.get(this);
        final EmbeddableEntity newEmbeddableEntity = (EmbeddableEntity) newValue;
        currentEmbeddableEntity.update(newEmbeddableEntity);
      } else {
        field.set(this, newValue);
      }
      this.dirtyFields.add(fieldName);
      this.updateTime();
      field.setAccessible(false);

      return true;

    } catch (final NoSuchFieldException e) {
      logger.error("Cannot find field", e);

    } catch (final IllegalAccessException e) {
      logger.error("Cannot access field", e);
    }

    return false;
  }

}
